import pytest

from ark import items, tools


def test_craft_without_subcomponents() -> None:
    available = {items.PASTE: 2000, items.METAL_INGOT: 1500}
    amount, plan, cost = tools.compute_crafting_plan(items.METAL_FOUNDATION, available)
    assert not plan
    assert amount == 30
    assert cost == {items.PASTE: 450, items.METAL_INGOT: 1500}


def test_craft_with_2_layers() -> None:

    available = {
        items.PASTE: 2000,
        items.METAL_INGOT: 1500,
        items.CRYSTAL: 150,
        items.SILICA_PEARL: 2250,
        items.ORGANIC_POLYMER: 500,
    }
    amount, plan, cost = tools.compute_crafting_plan(items.C4_DETONATOR, available)

    assert amount == 15
    assert plan == {items.ELECTRONICS: 750}
    assert cost == {
        items.PASTE: 225,
        items.CRYSTAL: 150,
        items.ELECTRONICS: 0,
        items.METAL_INGOT: 900,
        items.ORGANIC_POLYMER: 300,
        items.SILICA_PEARL: 2250,
    }


def test_craft_2layers_without_subcomponents() -> None:

    available = {
        items.PASTE: 2000,
        items.METAL_INGOT: 1500,
        items.CRYSTAL: 150,
        items.ELECTRONICS: 49,
        items.ORGANIC_POLYMER: 500,
    }
    amount, plan, cost = tools.compute_crafting_plan(items.C4_DETONATOR, available)
    assert not amount
    assert not plan
    assert not any(v for v in cost.values())


def test_raise_error_no_recipe_item() -> None:

    with pytest.raises(ValueError):
        tools.compute_crafting_plan(items.METAL_INGOT, {items.PASTE: 200})


def test_deep_layer_crafting() -> None:
    available = {
        items.METAL_INGOT: 2550,
        items.PASTE: 550,
        items.CRYSTAL: 200,
        items.ELECTRONICS: 70,
        items.SILICA_PEARL: 8000,
        items.ORGANIC_POLYMER: 10000,
        items.AUTO_TURRET: 1,
    }
    amount, plan, cost = tools.compute_crafting_plan(items.HEAVY_AUTO_TURRET, available)

    assert amount == 3
    assert plan == {items.ELECTRONICS: 670, items.AUTO_TURRET: 2}
    assert cost == {
        items.PASTE: 550,
        items.ELECTRONICS: 70,
        items.METAL_INGOT: 2150,
        items.ORGANIC_POLYMER: 190,
        items.SILICA_PEARL: 2010,
        items.AUTO_TURRET: 1,
    }

    available[items.AUTO_TURRET] = 0
    available[items.PASTE] = 600

    amount, plan, cost = tools.compute_crafting_plan(items.HEAVY_AUTO_TURRET, available)
    assert amount == 3
    assert plan == {items.ELECTRONICS: 740, items.AUTO_TURRET: 3}
    assert cost == {
        items.PASTE: 600,
        items.ELECTRONICS: 70,
        items.METAL_INGOT: 2360,
        items.ORGANIC_POLYMER: 210,
        items.SILICA_PEARL: 2220,
        items.AUTO_TURRET: 0,
    }


def test_amount_flattened() -> None:
    available = {
        items.SILICA_PEARL: 5211,
        items.PASTE: 2600,
        items.METAL_INGOT: 8757,
        items.ELECTRONICS: 1100,
        items.CRYSTAL: 10000,
        items.HIDE: 20000,
        items.ORGANIC_POLYMER: 10000,
    }

    amount, plan, cost = tools.compute_crafting_plan(items.HEAVY_AUTO_TURRET, available)
    assert amount == 10
    assert plan == {items.AUTO_TURRET: 10, items.ELECTRONICS: 1620}
    assert cost == {
        items.PASTE: 2000,
        items.ELECTRONICS: 1080,
        items.METAL_INGOT: 7020,
        items.ORGANIC_POLYMER: 700,
        items.SILICA_PEARL: 4860,
        items.AUTO_TURRET: 0,
    }

def test_tek_turret() -> None:
    available = {
        items.SILICA_PEARL: 5211,
        items.PASTE: 2600,
        items.METAL_INGOT: 8757,
        items.ELECTRONICS: 1100,
        items.CRYSTAL: 10000,
        items.HIDE: 20000,
        items.ORGANIC_POLYMER: 10000,
        items.ELEMENT: 18
    }
    amount, plan, cost = tools.compute_crafting_plan(items.TEK_TURRET, available)
    assert amount == 6
    assert not plan
    assert cost == {
        items.PASTE: 300,
        items.ELECTRONICS: 600,
        items.METAL_INGOT: 600,
        items.ORGANIC_POLYMER: 300,
        items.ELEMENT: 18
    }


from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Optional

from .._helpers import get_filepath


@dataclass
class Item:
    """Represents an ark item"""

    name: str
    search_name: str
    stack_size: int
    inventory_icon: str
    recipe: Optional[dict[Item, int]] = None
    added_icon: Optional[str] = None
    added_text: Optional[str] = None
    
    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return self.name

    def __post_init__(self) -> None:

        self.inventory_icon = get_filepath(self.inventory_icon)
        if self.added_icon is not None:
            self.added_icon = get_filepath(self.added_icon)

        if self.added_text is not None:
            self.added_text = get_filepath(self.added_text)

        for path in (self.inventory_icon, self.added_icon, self.added_text):
            if path is None:
                continue
            assert os.path.exists(path), f"Path not found: {path}"


from .item import Item

RIOT_HELMET = Item(
    name="Riot Helmet",
    search_name="helmet",
    stack_size=1,
    inventory_icon="assets/items/armor/riot_helmet.png",
)

RIOT_CHEST = Item(
    name="Riot Chest",
    search_name="chest",
    stack_size=1,
    inventory_icon="assets/items/armor/riot_chest.png",
)

RIOT_LEGGS = Item(
    name="Riot Leggings",
    search_name="legg",
    stack_size=1,
    inventory_icon="assets/items/armor/riot_leggs.png",
)

RIOT_GAUNTLETS = Item(
    name="Riot Gauntlets",
    search_name="gaunt",
    stack_size=1,
    inventory_icon="assets/items/armor/riot_gauntlets.png",
)

RIOT_BOOTS = Item(
    name="Riot Boots",
    search_name="boots",
    stack_size=1,
    inventory_icon="assets/items/armor/riot_boots.png",
)

MINER_HELMET = Item(
    name="Miner Helmet",
    search_name="miner helmet",
    stack_size=1,
    inventory_icon="assets/items/armor/miner_helmet.png",
)

RIOT = Item(
    name="Riot",
    search_name="riot",
    stack_size=1,
    inventory_icon="assets/items/armor/riot_chest.png",
)

TEK_GAUNTLETS = Item(
    name="Tek Gauntlets",
    search_name="tek gaunt",
    stack_size=1,
    inventory_icon="assets/items/armor/tek_gauntlets.png",
)

EXO_GLOVES = Item(
    name="Exo Gloves",
    search_name="exo gloves",
    stack_size=1,
    inventory_icon="assets/items/armor/exo_gauntlets.png",
)


import math
import random
from typing import Optional, final

import pyautogui as pg  # type: ignore[import]

from ... import config
from ..._helpers import await_event, get_center
from ...exceptions import (InventoryNotAccessibleError, MissingItemErrror,
                           NoItemsAddedError)
from ...items import Item
from .._button import Button
from .inventory import Inventory


@final
class PlayerInventory(Inventory):
    """Represents the player inventory in ark.

    Extends the abiliy of a regular `Inventory` by adding methods
    responsible for transferring items from the own inventory into
    a target inventory, as well as adjusting the region boundaries
    in order to apply the same methods to the player inventory that the
    `Inventory` already provides.
    """

    SLOTS = [
        (x, y, 93, 93) for y in range(232, 883, 93) for x in range(117, 582 + 93, 93)
    ]
    HEAD = (750, 155, 94, 94)
    TORSO = (750, 252, 94, 94)
    LEGS = (750, 349, 94, 94)
    HANDS = (1076, 155, 94, 94)
    OFFHAND = (1076, 252, 94, 94)
    FEET = (1076, 349, 94, 94)

    _CREATE_FOLDER = Button((513, 189))
    _TRANSFER_ALL = Button((350, 180))
    _DROP_ALL = Button((400, 180))
    _INVENTORY_TAB = Button((199, 117), (110, 90, 180, 60), "inventory.png")
    _CRAFTING_TAB = Button((1716, 118), (1627, 82, 211, 69), "crafting.png")

    _SEARCHBAR = (180, 180)
    _ADDED_REGION = (10, 1000, 220, 80)
    _ITEM_REGION = (117, 232, 582, 883)
    _UPPER_ITEM_REGION = (117, 230, 568, 191)
    _CAPPED_ICON = (85, 235, 52, 50)
    _YOU = (794, 116)

    def __init__(self):
        super().__init__("Player")

    def open(self, *_) -> None:
        """Opens the player inventory using the specified keybind.

        If the inventory did not open after 30 seconds,
        an `InventoryNotAccessibleError` is raised.
        """
        attempts = 0
        while not self.is_open():
            self.press(self.keybinds.inventory)
            if await_event(self.is_open, max_duration=5):
                return

            attempts += 1
            if attempts >= 6:
                raise InventoryNotAccessibleError(self._name)

    def await_items_added(self, item: Item | str) -> None:
        """Waits for items to be added to the inventory"""
        if not await_event(self.received_item, max_duration=30 * config.TIMER_FACTOR):
            raise NoItemsAddedError(item.name if isinstance(item, Item) else item)

    def transfer(
        self, item: Item, amount: int, target: Optional[Inventory] = None
    ) -> None:
        """Transfers an amount of an item into an inventory. The method used
        to transfer the items depends on the amount of stacks that has to be
        transferred, how many are already in the target inventory, and whether
        a target inventory was given to begin with.

        If more than 40 stacks of items need to be transferred, additionally
        to what is already within the target, or no target inventory was given,
        it will simply iterate over the top row of items and OCR the amount
        of items removed.

        Otherwise, we can check on the amount of stacks in the target before
        transferring and compare them to after, for each item. This way we can
        accurately count the items transferred while being more lag proof.

        Parameters:
        -----------
        item :class:`str`:
            The item to search for before transferring

        amount :class:`int`:
            The quantity of items to be transferred

        target_inventory :class:`Inventory`: [Optional]
            The inventory to transfer the items to, required for stack transferring
        """
        self.search(item)
        # round the amount to the next stacksize, i.e if 403 items are to
        # be transferred, and the stacksize is 200, it will be rounded to 600
        amount = int(math.ceil(amount / item.stack_size)) * item.stack_size
        stacks = int(amount / item.stack_size)
        rows = round(stacks / 6)

        if rows > 7 or not target:
            self._transfer_by_rows(item, rows)
            return

        # search the item in the target so we can track how much we already
        # transferred. Before that check that it actually has the slots free
        # we need, if not we need to go back to doing it by row.
        target.search(item)
        self.sleep(0.3)

        stacks_in_target = target.count(item)
        if (stacks_in_target + stacks) > 42:
            self._transfer_by_rows(item, rows)
            return

        if stacks_in_target >= stacks:
            return

        self._transfer_by_stacks(item, stacks, target)

    def equip(self, item: Item) -> None:
        self.search(item)
        self.sleep(0.3)

        pos = self.find(item)
        if pos is None:
            raise MissingItemErrror(item.name)

        before = self.count(item)
        self.click_at(self._YOU, delay=0.2)
        self.sleep(1)
        self.click_at(pos)
        self.press(self.keybinds.use)

        if not await_event(
            lambda: before != self.count(item),
            max_duration=30 * config.TIMER_FACTOR,
            ignore_annotation=True,
        ):
            raise TimeoutError

    def unequip(self, item: Item) -> None:
        self.click_at(self._YOU, delay=0.2)
        for slot in [
            self.HEAD,
            self.TORSO,
            self.LEGS,
            self.HANDS,
            self.OFFHAND,
            self.FEET,
        ]:
            if not self._slot_has_item(slot, item):
                continue
            self.click_at(get_center(slot))
            self.press(self.keybinds.use)

            if not await_event(
                lambda: self._slot_has_item(slot, item),
                False,
                max_duration=30 * config.TIMER_FACTOR,
                ignore_annotation=True,
            ):
                raise TimeoutError
            return
        raise MissingItemErrror(item.name)

    def transfer_spam(self, item: Item, times: int) -> None:
        self.search(item)

        for _ in range(times):
            slot = random.choice(self.SLOTS)
            self.move_to(get_center(slot))
            self.press(self.keybinds.transfer)

    def hp_full(self) -> bool:
        return pg.pixelMatchesColor(*(1118, 514), (15, 166, 181), tolerance=40)

    def food_full(self) -> bool:
        return pg.pixelMatchesColor(*(1118, 643), (15, 166, 181), tolerance=40)

    def water_full(self) -> bool:
        return pg.pixelMatchesColor(*(1118, 685), (15, 166, 181), tolerance=40)

    def _slot_has_item(self, slot: tuple[int, int, int, int], item: Item) -> bool:
        return (
            self.window.locate_template(
                item.inventory_icon, slot, confidence=0.7, grayscale=True
            )
            is not None
        )

    def _transfer_by_stacks(self, item: Item, stacks: int, target: Inventory) -> None:
        """Internal implementation of the stack transferring technique.

        Counts the amount of items before pressing 'T' and waits for it to
        change after the press.
        """
        amount = item.stack_size * stacks

        for _ in range(stacks):
            slot = self.find(item)
            if slot is None:
                return

            pg.moveTo(slot)
            before = target.count(item)
            self.press(self.keybinds.transfer)
            try:
                target._receive_stack(item, before)
            except NoItemsAddedError:
                continue

            transferred = target.count(item) * item.stack_size
            print(f"Transferred {transferred}/{amount}...")
            if amount <= transferred <= amount + 3000:
                return

    def _transfer_by_rows(self, item: Item, rows: int) -> None:
        """Internal implementation of the row transferring technique.

        OCRs the amount transferred after each row
        """
        EXTRA_ITERATION_FACTOR = 1.5
        amount = item.stack_size * (rows * 6)

        for _ in range(round(rows * EXTRA_ITERATION_FACTOR) * 6):
            slot = self.find(item)
            if slot is None:
                return

            pg.moveTo(slot)
            self.press(self.keybinds.transfer)
            transferred = self.get_amount_transferred(item, "rm")
            if not transferred:
                continue

            print(f"Transferred {transferred}/{amount}...")
            if amount <= transferred <= (amount + 10 * item.stack_size):
                return



from .entities import *
from .interfaces import *
from .server import Server
from .settings import InputSettings, UserSettings, DinoExport
from .state import State
from .window import ArkWindow

__all__ = ("State", "ArkWindow", "Server", "InputSettings", "UserSettings", "DinoExport")
__version__ = "1.2.4"



import pyautogui  # type:ignore[import]
from pytesseract import pytesseract as tes  # type:ignore[import]

from .._ark import Ark
from .._helpers import await_event
from ..exceptions import InterfaceError, TimerNotVisibleError


class HUDInfo(Ark):
    """Represents the extended HUD Info in Ark.
    
    Provides the ability to get the current timer assuming tek
    gauntlets are equipped.
    """

    _DAY_REGION = (0, 10, 214, 95)
    _TIMER_WORD_REGION = (90, 130, 78, 28)
    _TIMER_REGION = (164, 127, 69, 30)

    def open(self) -> None:
        """Opens the HUD info, key stays in a held state!"""
        attempts = 0
        while not self.is_open():
            pyautogui.keyDown(self.keybinds.hud_info)

            if await_event(self.is_open, max_duration=3):
                return
            pyautogui.keyUp(self.keybinds.hud_info)
            self.sleep(0.3)
            attempts += 1
            if attempts > 4:
                raise InterfaceError("Failed to open HUD!")

    def is_open(self) -> bool:
        """Returns whether the HUD info interface is open."""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/day.png",
                region=self._DAY_REGION,
                confidence=0.75
            )
            is not None
        )

    def can_get_timer(self) -> bool:
        """Returns whether the timer is visible"""
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/interfaces/timer.png",
                region=self._TIMER_WORD_REGION,
                confidence=0.7
            )
            is not None
        )

    def get_timer(self) -> int | None:
        """Returns the timer as total seconds, or `None` if not determined.
        
        If the timer is not visible, it will raise a `TimerNotVisibleError`.
        """
        self.open()
        if not self.can_get_timer():
            raise TimerNotVisibleError
        return self._fetch_timer()

    def _fetch_timer(self) -> int | None:
        """Fetches the timer by denoising the timer region and then OCR'ing
        the text."""
        img = self.window.grab_screen(self._TIMER_REGION)
        img = self.window.denoise_text(img, (63, 179, 255), variance=15)

        raw: str = tes.image_to_string(
            img, config=f"-c tessedit_char_whitelist=0123456789liIxObL: --psm 7 -l eng"
        ).strip()

        if ":" not in raw or len(raw) < 4:
            return None
            
        minutes, seconds = raw.split(":")
        return (int(minutes) * 60) + int(seconds)



import cv2 as cv  # type: ignore[import]
import numpy as np
import pydirectinput as input  # type: ignore[import]

from .dinosaur import Dinosaur


class Stryder(Dinosaur):
    """Represents a stryder in ark, used for resource logistics."""

    def __init__(self) -> None:
        super().__init__("Tek Stryder", "assets/wheels/stryder.png")

    def find_dedi_transfer_tab(self) -> tuple | None:
        """Finds the dedi transfer tab by denoising for the orange text.

        Matching for the orange text template is not reliable because the text
        size changes depending on how many options are on the action wheel.

        Returns the coordinate of the transfer tab as a tuple, or None if not found.
        """
        # grab action wheel region, denoise for orange text
        img = self.window.grab_screen((470, 80, 980, 900))
        mask = self.window.denoise_text(img, (39, 146, 255), 15)

        # get the first orange matching pixel
        matches = cv.findNonZero(mask)
        if matches is None:
            return None

        return [(470 + x, 80 + y) for x, y in np.concatenate(matches, axis=0)][0]

    def open_dedi_transfer_wheel(self) -> None:
        """Opens the dedi transfer action wheel."""
        # find position of the wheel option
        pos = self.find_dedi_transfer_tab()
        if not pos:
            raise LookupError("Failed to find dedi transfer action!")

        # enter the option
        self.sleep(0.5)
        input.moveTo(*pos, duration=0.1)
        self.sleep(0.1)
        input.moveTo(*pos, duration=0.1)
        self.sleep(1)
        self.click("left")
        self.sleep(1)

    def click_sort_to_nearby_dedis(self) -> None:
        """Clicks the 'sort to nearby dedis' in the open action wheel."""
        input.keyDown("e")
        input.moveTo(960, 230, duration=0.1)
        self.sleep(0.1)
        input.moveTo(960, 230, duration=0.1)
        self.sleep(1)
        self.click("left")
        input.keyUp("e")

    def sort_items_to_nearby_dedis(self) -> None:
        """Uses the stryders action wheel to sort all the items in the stryders
        inventory to nearby dedis."""
        self.action_wheel.activate()
        self.sleep(2)
        self.open_dedi_transfer_wheel()
        self.click_sort_to_nearby_dedis()


import os
from dataclasses import dataclass

from .._helpers import get_filepath


@dataclass
class Buff:
    name: str
    image: str


    def __post_init__(self) -> None:
        self.image = get_filepath(self.image)

        assert os.path.exists(self.image), f"Path not found: {self.image}"


from typing import Any, Optional

import cv2  # type: ignore[import]
from pytesseract import pytesseract as tes  # type: ignore[import]

from ..._ark import Ark
from ...exceptions import InventoryNotAccessibleError, NoGasolineError, WheelError
from ...items import Item
from .._button import Button
from ..inventories import Inventory
from ..wheels import ActionWheel


class Structure(Ark):
    """Represents a structure in ARK.

    A structure provides access to it's inventory and allows to toggle
    it on or off, as well as check how many items were deposited into it
    and whats currently inside of it. The `name` does not have to match
    it's real in-game name, you may alter or enumerate the station names
    to keep track of different objects.

    Parameters
    ----------
    name :class:`str`:
        The name of the structure it represents.

    action_wheel :class:`ActionWheel | str`:
        The action wheel of the structure, either a previously created wheel
        or the filepath to the wheel image to create the wheel.

    inventory :class:`Inventory` [Optional]:
        The pre-created inventory of the structure, if not passed one will be created.

    craftables :class:`list[Item]` [Optional]:
        A list of items that can be crafted in the structure, onl to be passed if
        `inventory` was not.

    capacity :class:`str | int` [Optional]:
        The capacity of the structure, either as integer or a filepath leading
        to the file of the capped structure.

    toggleable :class:`bool`:
        Whether the structure can be turned on and off, `False` by default.

    Attributes:
    -----------
    TURN_ON :class:`Button`-
        A representation of the 'turn on' button for toggleable structures

    TURN_OFF :class:`Button`-
        A representation of the 'turn off' button for toggleable structures

    Properties:
    -----------
    name :class:`str`[get]:
        The name the structure was initialized with

    toggleable :class:`bool`[get]:
        Whether the structure is toggleable
    """

    TURN_ON = Button((956, 618), (740, 570, 444, 140), "turn_on.png")
    TURN_OFF = Button((956, 618), (740, 570, 444, 140), "turn_off.png")
    _ITEM_ADDED_REGION = (5, 850, 55, 230)

    def __init__(
        self,
        name: str,
        action_wheel: ActionWheel | str,
        inventory: Optional[Inventory] = None,
        craftables: Optional[list[Item]] = None,
        capacity: Optional[int | str] = None,
        *,
        toggleable: bool = False,
    ) -> None:
        super().__init__()
        if inventory and any((craftables, capacity)):
            raise ValueError(
                "Did not expect 'craftables' or 'capacity' alongside 'inventory'."
            )
        if isinstance(action_wheel, str):
            action_wheel = ActionWheel(name, action_wheel)
        if inventory is None:
            self.inventory = Inventory(name, craftables, capacity)
        else:
            self.inventory = inventory
        self.action_wheel = action_wheel
        self._name = name
        self._toggleable = toggleable

    def __str__(self) -> str:
        return f"Structure '{self._name}' of type '{type(self).__name__}'"

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(name={self._name}, inventory={self.inventory},"
            f"action_wheel={self.action_wheel}, toggleable={self._toggleable})"
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def toggleable(self) -> bool:
        return self._toggleable

    def open(self) -> None:
        """Wraps the inventory `open` function using the action wheel to
        validate whether we are actually in range of it on failure.

        Raises
        ------
        `InventoryNotAccessibleError`
            If the Inventory could not be accessed even though it is in
            range (determined by the action wheel)

        `WheelNotAccessibleError`
            When no wheel could be opened at all after several attempts

        `UnexpectedWheelError`
            When a wheel was opened but is of an unexpected entity
        """
        try:
            self.inventory.open()
        except InventoryNotAccessibleError as e:
            try:
                self.action_wheel.activate()
            except WheelError:
                self.action_wheel.deactivate()
                raise e
            self.action_wheel.deactivate()
            self.inventory.open(max_duration=40)

    def close(self) -> None:
        """Wraps the inventory `close` function."""
        self.inventory.close()

    def _get_amount_deposited(self, item: Item) -> int:
        """Checks how much of the given item was deposited. This is achieved
        by locating the items `added_icon` on the left side of the screen,
        and then doing some processing to get a region of interest of the
        amount that has been deposited.

        The image is then denoised, OCR'd and validated.

        Parameters
        ----------
        item :class:`Item`:
            The item to determine the amount of

        Returns
        -------
        :class:`int`:
            The amount of dust deposited, 0 on failure

        Example usage:
        ```py
        self._get_amount_deposited(DUST)
        >>> 2376
        ```
        """
        roi = self._get_item_amount_roi(item)
        if roi is None:
            return 0

        # grab the region of interest and apply denoising
        img = self.window.grab_screen(roi, convert=False)
        img = self.window.denoise_text(img, denoise_rgb=(255, 255, 255), variance=10)
        cv2.imshow("", img)
        cv2.waitKey(1)
        
        raw_result = tes.image_to_string(
            img,
            config="-c tessedit_char_whitelist=0123456789liIxObL --psm 7 -l eng",
        )
        return int(self._correct_ocr_mistakes(raw_result))

    def _get_item_amount_roi(self, item: Item) -> tuple[int, int, int, int] | None:
        """Returns the region of interest of a deposited item to determine the
        amount that was deposited.

        Parameters
        ----------
        item :class:`Item`:
            The item to determine the region of interest of

        Returns
        -------
        :class:`tuple[int, int, int, int]` | `None`:
            The region of interest as `x, y, w, h` or `None` on failure

        """
        if item.added_text is None:
            raise ValueError(f"Undefined 'added_text' for {item}!")

        if not (icon_pos := self._find_item_deposited(item)):
            return None

        # compute name roi
        text_start_x = icon_pos[0] + icon_pos[2]
        name_roi = (text_start_x + 130, icon_pos[1], 290, 40)
        name_text = self.window.locate_template(
            item.added_text,
            region=name_roi,
            confidence=0.7,
            convert=False,
        )
        # not worth trying to find out the amount without proper roi
        if name_text is None:
            return None

        # get our region of interest (from end of "Removed:" to start of "Element")
        return (name_roi[0], name_roi[1], int(name_text[0] - name_roi[0]), 40)

    def _correct_ocr_mistakes(self, raw: str) -> str:
        """Corrects 'expected' or common tesseract mistakes."""
        # replace all the mistaken "1"s
        for char in ["I", "l", "i", "b", "L"]:
            raw = raw.replace(char, "1")

        # replace mistaken "0"s, strip off newlines
        filtered = raw.replace("O", "0").rstrip()

        # find the x to slice out the actual number
        x = filtered.find("x")
        if any((not filtered, x == -1, filtered == "x")):
            return "0"

        return filtered[:x]

    def _find_item_deposited(self, item: Item) -> Any:
        """Returns the position of a deposited item. The `Item` must have a
        `added_icon` defined, otherwise a `ValueError` is raised."""
        if item.added_icon is None:
            raise ValueError(f"Undefined 'added_icon' for {item}!")

        return self.window.locate_template(
            item.added_icon, region=self._ITEM_ADDED_REGION, confidence=0.7
        )

    def _find_turn_on_text(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/turn_on.png",
                (0, 0, 1920, 1080),
                confidence=0.92,
            )
            is not None
        )

    def _find_turn_off_text(self) -> bool:
        return (
            self.window.locate_template(
                f"{self.PKG_DIR}/assets/templates/turn_off.png",
                (0, 0, 1920, 1080),
                confidence=0.92,
            )
            is not None
        )

    def turn_on(self) -> None:
        """Turns the structure on, if not inside an inventory, it will first
        look for the 'TURN ON' text to toggle the structure from outside.

        Otherwise the inventory will be opened (if it is not already), and
        the structure will be toggled through the inventory.

        Raises a `NoGasolineError` if the structure cannot be turned on.
        """
        if not self.inventory.is_open() and self._find_turn_on_text():
            self.press(self.keybinds.use)
            self.sleep(0.3)
            return

        self.inventory.open()
        if self.is_turned_on():
            return

        if not self.is_turned_off():
            raise NoGasolineError(self._name)

        while self.is_turned_off():
            self.click_at(964, 615, delay=0.3)
            self.sleep(1)

    def turn_off(self) -> None:
        """Turns the structure off, assumes it is already opened."""

        if not self.inventory.is_open() and self._find_turn_off_text():
            self.press(self.keybinds.use)
            self.sleep(0.3)
            return

        self.inventory.open()
        if self.is_turned_off():
            return

        while self.is_turned_on():
            self.click_at(964, 615, delay=0.3)
            self.sleep(1)

    def is_turned_on(self) -> bool:
        """Return whether the Structure is turned on."""
        if not self._toggleable:
            raise ValueError(f"{self} is not toggleable!")
        return self.inventory.locate_button(
            self.TURN_OFF, grayscale=True, confidence=0.85
        )

    def is_turned_off(self) -> bool:
        """Return whether the Structure can be turned on."""
        return self.inventory.locate_button(
            self.TURN_ON, grayscale=True, confidence=0.85
        )



class State:
    """Controls the state of the program."""

    running = True
    paused = False

    def __init__(self) -> None:
        raise RuntimeError("State class is not meant to be instantiated")



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class Stats:
    health: int
    food: int
    water: int
    weight: int
    stamina: Optional[int] = None
    oxygen: Optional[int] = None
    melee: Optional[int] = None
    speed: Optional[int] = None
    crafting: Optional[int] = None
    fortitude: Optional[int] = None


from dataclasses import dataclass
from typing import Optional
from .._ark import Ark
from .._helpers import get_filepath

@dataclass
class Button:

    location: tuple[int, int]
    region: Optional[tuple[int, int, int, int]] = None
    template: Optional[str] = None

    def __post_init__(self) -> None:
        if self.template is None:
            return
        template = f"{Ark.PKG_DIR}/assets/interfaces/{self.template}"
        self.template = get_filepath(template)


import pyautogui  # type:ignore[import]

from .wheel import ActionWheel


class TekPodWheel(ActionWheel):
    def __init__(self) -> None:
        super().__init__("Tek Sleeping Pod", "assets/wheels/pod.png")

    def lay_on(self) -> None:
        self.select_action((1166, 495), click=False)
        pyautogui.keyUp(self.keybinds.use)



from .item import Item

MEJOBERRY = Item(
    name="Mejoberry",
    search_name="mejoberry",
    stack_size=100,
    inventory_icon="assets/items/consumables/mejoberries.png",
)

NARCOBERRY = Item(
    name="Narcoberry",
    search_name="narcoberry",
    stack_size=100,
    inventory_icon="assets/items/consumables/narcoberries.png",
)

STIMBERRY = Item(
    name="Stimberry",
    search_name="stimberry",
    stack_size=100,
    inventory_icon="assets/items/consumables/stimberries.png",
)

TINTOBERRY = Item(
    name="Tintoberry",
    search_name="tintoberry",
    stack_size=100,
    inventory_icon="assets/items/consumables/tintoberries.png",
)

AMARBERRY = Item(
    name="Amarberry",
    search_name="amarberry",
    stack_size=100,
    inventory_icon="assets/items/consumables/amarberries.png",
)

AZULBERRY = Item(
    name="Azulberry",
    search_name="azulberry",
    stack_size=100,
    inventory_icon="assets/items/consumables/azulberries.png",
)

RAW_MEAT = Item(
    name="Raw Meat",
    search_name="meat",
    stack_size=40,
    inventory_icon="assets/items/consumables/raw_meat.png",
)

RAW_FISH_MEAT = Item(
    name="Raw Fish Meat",
    search_name="fish",
    stack_size=40,
    inventory_icon="assets/items/consumables/raw_fish.png",
)

SPOILED_MEAT = Item(
    name="Spoiled Meat",
    search_name="spoiled",
    stack_size=100,
    inventory_icon="assets/items/consumables/spoiled_meat.png",
)

PELLET = Item(
    name="Snow Owl Pellet",
    search_name="pellet",
    stack_size=1,
    inventory_icon="assets/items/consumables/pellet.png",
)


YTRAP_SEED = Item(
    name="Plant Species Y Seed",
    search_name="y seed",
    stack_size=100,
    inventory_icon="assets/items/consumables/ytrap_seed.png",
)